/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998 Benjamin Geer
 * Email: beroul@yahoo.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.template;

import java.util.*;
import java.util.EventObject;
import java.util.EventListener;
import java.io.*;

/**
 * A <tt>TemplateCache</tt> that loads templates from a filesystem.  Given a directory
 * path, the cache assumes by default that all files in the directory are templates.  It
 * can optionally be given a filename suffix for templates.  Before using a
 * <tt>FileTemplateCache</tt>, you must call its <tt>startAutoUpdate()</tt> method to
 * begin periodic updates.  The default loading policy is LOAD_ON_DEMAND: templates are
 * loaded into the cache only when requested, each template's modification date is checked
 * each time it is requested, and the periodic updates are used only to remove deleted
 * templates from the cache.  If the loading policy is set to PRELOAD, all templates are
 * loaded when <tt>startAutoUpdate()</tt> is called, and all files are checked during each
 * periodic update.  If template files will not be changed frequently, use LOAD_ON_DEMAND
 * with a long delay value for maximum performance.
 *
 * <p>The string argument to the <tt>getTemplate()</tt> method is interpreted as the
 * template's path relative to the cache's root directory, using a forward slash
 * (<tt>/</tt>) as a separator (this is to facilitate using URL path info to request
 * templates).  For example, if a <tt>TemplateCache</tt> object was made for the directory
 * <tt>templates</tt>, which contains a subdirectory <tt>foo</tt>, in which there is a
 * template file called <tt>index.html</tt>, you would call
 * <tt>getTemplate("foo/index.html")</tt> to retrieve that template.
 *
 * <p>The owner of the cache should implement <tt>CacheListener</tt>, and
 * register itself using <tt>addCacheListener()</tt>.
 *
 * <p>If the template cannot read its cache directory, the periodic updates will be
 * cancelled until the next time <tt>startAutoUpdate()</tt> is called.
 *
 * @see TemplateCache
 * @see CacheListener
 */
public class FileTemplateCache implements TemplateCache, Updatable {

	/**
	 * Used with <tt>setLoadingPolicy()</tt> to indicate that templates
	 * should be loaded as they are requested.
	 */
	public static final int LOAD_ON_DEMAND = 0;

	/**
	 * Used with <tt>setLoadingPolicy()</tt> to indicate that templates
	 * should be preloaded.
	 */
	public static final int PRELOAD = 1;

    private File cacheDir;
    private Map cache = new HashMap();
    private UpdateTimer timer;
    private long delay = 5000; // five seconds
	private int loadingPolicy = LOAD_ON_DEMAND;
    private GenericEventMulticaster multicaster = new GenericEventMulticaster();
	private String encoding;

	// A ListenerAdapter for cacheUnavailable events.
    private ListenerAdapter unavailableAdapter = new ListenerAdapter() {
	    public void fireEvent(EventObject event, EventListener listener) {
			((CacheListener)listener).cacheUnavailable((CacheEvent)event);
	    }
	};

	// A ListenerAdapter for elementUpdated events.
    private ListenerAdapter updatedAdapter = new ListenerAdapter() {
		public void fireEvent(EventObject event, EventListener listener) {
			((CacheListener)listener).elementUpdated((CacheEvent)event);
		}
	};

	// A ListenerAdapter for elementUpdateFailed events.
    private ListenerAdapter updateFailedAdapter = new ListenerAdapter() {
		public void fireEvent(EventObject event, EventListener listener) {
			((CacheListener)listener).elementUpdateFailed((CacheEvent)event);
		}
	};

	// A ListenerAdapter for elementRemoved events.
    private ListenerAdapter removedAdapter = new ListenerAdapter() {
		public void fireEvent(EventObject event, EventListener listener) {
			((CacheListener)listener).elementRemoved((CacheEvent)event);
		}
	};

    private String filenameSuffix;

    /**
     * Constructs an empty <tt>FileTemplateCache</tt>.
     */
    public FileTemplateCache() { }

    /**
     * Constructs a <tt>TemplateCache</tt> with a directory in which
     * it will look for template files.
     *
     * @param path the absolute path of the directory containing
     * templates for this cache.
     */
    public FileTemplateCache(String path) {
		setPath(path);
    }

    /**
     * Constructs a <tt>TemplateCache</tt> with a directory in which
     * it will look for template files.
     *
     * @param dir the directory containing templates for this cache.
     */
    public FileTemplateCache(File dir) {
		setDirectory(dir);
    }

    /**
     * Constructs a <tt>TemplateCache</tt> with a directory in which
     * it will look for template files, and a delay representing the
     * number of seconds between cache updates.
     *
     * @param path the absolute path of the directory containing
     * templates for this cache.
     * @param delay the number of seconds between cache updates.
     */
    public FileTemplateCache(String path, long delay) {
		setPath(path);
		setDelay(delay);
    }

    /**
     * Constructs a <tt>TemplateCache</tt> with a directory in which
     * it will look for template files, and a delay representing the
     * number of seconds between cache updates.
     *
     * @param dir the directory containing templates for this cache.
     * @param delay the number of seconds between cache updates.
     */
    public FileTemplateCache(File dir, long delay) {
		setDirectory(dir);
		setDelay(delay);
    }

	/**
	 * Sets the loading policy for this <tt>FileTemplateCache</tt>.  If
	 * <tt>LOAD_ON_DEMAND</tt>, templates will be loaded as they are requested, and each
	 * template's file modification date will be checked each time it is requested.  If
	 * <tt>PRELOAD</tt>, all templates in the cache directory and its subdirectories will
	 * be loaded when the cache is started, and new templates will be added to the cache
	 * each time it is updated.  Defaults to <tt>LOAD_ON_DEMAND</tt>.
	 */
	public void setLoadingPolicy(int loadingPolicy) {
		this.loadingPolicy = loadingPolicy;
	}

    /**
     * @param path the absolute path of the directory containing
     * templates for this cache.
     */
    public void setPath(String path) {
		setDirectory(new File(path));
    }

    /**
     * @param dir the directory containing templates for this cache.
     */
    public synchronized void setDirectory(File dir) {
		cache.clear();
		this.cacheDir = dir;
    }

    /**
     * @param delay the number of seconds between cache updates.
     * Defaults to five seconds.
     */
    public synchronized void setDelay(long delay) {
		this.delay = delay * 1000;
    }

	/**
	 * Sets the character encoding to be used when reading template files.
	 *
	 * @param encoding the name of the encoding to be used; this will be passed to the
	 * constructor of <tt>InputStreamReader</tt>.
	 */
	public void setEncoding(String encoding) {
		this.encoding = encoding;
	}

    /**
     * @param filenameSuffix the optional filename suffix of template
     * files to be read for this cache.  If this suffix is set,
     * files that do not have the suffix will be ignored.
     */
    public void setFilenameSuffix(String filenameSuffix) {
		this.filenameSuffix = filenameSuffix;
    }

    public void addCacheListener(CacheListener listener) {
		multicaster.addListener(listener);
    }

    public void removeCacheListener(CacheListener listener) {
		multicaster.removeListener(listener);
    }

    /**
     * Gets a template from the cache.
     *
     * @param name the template's filename, including its
     * path relative to the cache's root directory.
     * @return the corresponding <tt>Template</tt>, or null
     * if not found.
     */
    public synchronized Template getTemplate(String name) {
		if (!(filenameSuffix == null || name.endsWith(filenameSuffix))) {
			fireCacheEvent(updateFailedAdapter, name,
						   new IOException("The requested name, \"" + name +
										   "\", does not have the filename suffix \"" +
										   filenameSuffix + "\"."));
			return null;
		}
		
		// If we're preloading, then we can just return the template from the cache
		// if we have it; it will be updated the next time our update() method
		// is called.
		if (loadingPolicy == PRELOAD) {
			CacheElement element = (CacheElement)cache.get(name);
			if (element != null) {
				return element.template;
			} else {
				return null;
			}
		} else {
			// If we're loading on demand, check the file.
			if (!checkCacheDir()) return null;
			
			File file = nameToFile(name);
			if (!file.exists()) return null;
			
			if (!file.isFile()) {
				fireCacheEvent(updateFailedAdapter,
							   name,
							   new IOException("\"" +
											   file.getAbsolutePath() +
											   "\" is a directory."));
				return null;
			}
			
			return getLatestTemplate(name, file);
		}
	}

    public void startAutoUpdate() {
		synchronized (this) {
			if (timer == null) {
				timer = new UpdateTimer(this, delay);
			}
			update();
			timer.startTiming();
		}
    }

    public void stopAutoUpdate() {
		if (timer != null) {
			timer.stopTiming();
		}
    }

	/**
	 * @return true if the cache directory exists and is readable;
	 * otherwise, also fires an appropriate <tt>CacheEvent</tt>.
	 */
	private boolean checkCacheDir() {
		if (!cacheDir.exists()) {
			fireCacheEvent(unavailableAdapter, null, new IOException("Directory \"" +
				cacheDir.getAbsolutePath() + "\" not found."));
			stopAutoUpdate();
			return false;
		}

		if (!cacheDir.isDirectory()) {
			fireCacheEvent(unavailableAdapter, null,
						   new IOException("\"" + cacheDir.getAbsolutePath() +
				"\" is not a directory."));
			stopAutoUpdate();
			return false;
		}

		return true;
	}

	private Template loadTemplate(File file) throws IOException {
		if (encoding == null) {
			return new Template(file);
		} else {
			return new Template(new InputStreamReader(new FileInputStream(file),
													  encoding));
		}
	}

	/**
	 * Given a file, returns the template from the file if it's more
	 * recent then the one in the cache; otherwise, returns the one
	 * in the cache.
	 *
	 * @param elementName the name the template should have in the cache.
	 * @param file the file from which the template may be read.
	 */
	private Template getLatestTemplate(String elementName, File file) {
		Template template = null;
		long lastModified = file.lastModified();
		CacheElement element = (CacheElement)cache.get(elementName);

		try {
			// If we don't have this template, add it.
			if (element == null) {
				template = loadTemplate(file);
				template.setTemplateCache(this);
				cache.put(elementName, new CacheElement(template, lastModified));
				fireCacheEvent(updatedAdapter, elementName, null);

			// If we have it and it's been modified, update it.
			} else if (lastModified > element.lastModified) {
				template = loadTemplate(file);
				template.setTemplateCache(this);
				element.template = template;
				element.lastModified = lastModified;
				fireCacheEvent(updatedAdapter, elementName, null);

			} else {
				// Return the Template object we already have.
				template = element.template;
			}
		} catch (IOException e) {
			fireCacheEvent(updateFailedAdapter, elementName, e);
		}

		return template;
	}

    /**
     * Updates the cache.
     */
    public synchronized void update() {
		if (!checkCacheDir()) return;

		switch (loadingPolicy) {
		case LOAD_ON_DEMAND:
		default:
			removeDeletedTemplates();
			break;
		case PRELOAD:
			LinkedList visitedFiles = new LinkedList();
			try {
				readDirectory(cacheDir, "", visitedFiles);
			} catch (IOException e) {
				stopAutoUpdate();
				fireCacheEvent(unavailableAdapter, null, e);
				return;
			}

			// Remove from the cache any templates corresponding to files we
			// didn't just visit.
			Set keySet = cache.keySet();
			Iterator keyIterator = keySet.iterator();
			while (keyIterator.hasNext()) {
				String elementName = (String)keyIterator.next();
				if (!visitedFiles.contains(elementName)) {
					keyIterator.remove();
					fireCacheEvent(removedAdapter, elementName, null);
				}
			}
			break;
		}
    }

	/**
	 * Removes from the cache templates that correspond to deleted files.
	 */
	private void removeDeletedTemplates() {
		Set keySet = cache.keySet();
		Iterator keyIterator = keySet.iterator();
		while (keyIterator.hasNext()) {
			String elementName = (String)keyIterator.next();
			File file = nameToFile(elementName);
			if (!file.exists()) {
				keyIterator.remove();
				fireCacheEvent(removedAdapter, elementName, null);
			}
		}
	}		

    /**
     * Recursively updates the cache from the files in a (sub)directory
	 * and its subdirectories.
     *
     * @param dir the directory to be read.
     * @param relativeDirPath a string representing the directory's path
     * relative to the root cache directory.
	 * @param visitedFiles a List of files that have been visited so far.
     */
    private void readDirectory(File dir, String relativeDirPath, List visitedFiles)
		throws IOException {
		String[] filenames = dir.list();

		if (filenames == null) {
			throw new IOException("Could not get file list from directory \"" +
				dir.getAbsolutePath() + "\".");
		}

		// Iterate through the items in the directory.
		for (int fileNum = 0; fileNum < filenames.length; fileNum++) {
			String filename = filenames[fileNum];
			File file = new File(dir, filename);
			String elementName = relativeDirPath + filename;

			// If the item is a file, see if we need to to read it.
			if (file.isFile()) {

				// If we have no filename suffix, or if we have one and this
				// file ends with it, check the file.
				if (filenameSuffix == null || filename.endsWith(filenameSuffix)) {
					visitedFiles.add(elementName);
					getLatestTemplate(elementName, file);
				}
			} else if (file.isDirectory()) {
				// If the item is a directory, recursively read it.
				readDirectory(file, elementName + "/", visitedFiles);
			}
		}
    }

	/**
	 * Converts a cache element name to a <tt>File</tt>.
	 */
	private File nameToFile(String name) {
		// Replace forward slashes with the operating system's
		// file separator, if it's not a forward slash.
		if (File.separatorChar == '/') {
			return new File(cacheDir, name);
		} else {
			StringBuffer pathBuf = new StringBuffer();

			for (int i = 0; i < name.length(); i++) {
				char c = name.charAt(i);
				if (c == '/') {
					pathBuf.append(File.separatorChar);
				} else {
					pathBuf.append(c);
				}
			}

			return new File(cacheDir, pathBuf.toString());
		}
	}

	/**
	 * A convenience method for firing a CacheEvent.
	 *
	 * @param adapter a <tt>ListenerAdapter</tt>.
	 * @param elementName the name of the cache element in question, or null.
	 * @param e an <tt>Exception</tt> to be included with the event, or null.
	 */
	private void fireCacheEvent(ListenerAdapter adapter, String elementName, Exception e) {
		CacheEvent event = new CacheEvent(this);
		event.setElementName(elementName);
		event.setException(e);
		multicaster.fireEvent(event, adapter);
	}

    /**
     * A template and its modification date.
     */
    private class CacheElement {
		Template template;
		long lastModified;

		CacheElement(Template template, long lastModified) {
			this.template = template;
			this.lastModified = lastModified;
		}
    }
}
